{.$DEFINE Debug}

{$IFDEF Debug}

  {$M $800, 0, $2000}

{$ELSE}

  {$M $400, 0, $2000}

{$ENDIF}

program
  PAFRT;

uses
  Dos,
  Objects,

  RTC,
  BiosDisk;

type
  PFAT = ^TFAT;
  TFAT = record
    Desc: Byte;
    Dummy: Word;
  end;

  PDirEntry = ^TDirEntry;
  TDirEntry = record
    Name: array [1..8] of Char;
    Ext: array [1..3] of Char;
    Attr: Byte;
    Reserved: array [1..$A] of Byte;
    Time: Word;
    Date: Word;
    Cluster: Word;
    Size: Longint;
  end;

  PSector = ^TSector;
  TSector = array [0..$200-1] of Byte{s};

var
  Disk: array [0..25-1] of TSector;
  DiskBytes: array [0..25*SizeOf (TSector)-1] of Byte absolute Disk;
const
  StartData = 25;
var
  TotalClusters: Word;
  RTStartData: Word;
  RTSegments: Word;

procedure BootSec; external; {$L BootSec}

procedure InitDisk;
const
  FAT: TFAT = (
    Desc: $F0;
    Dummy: $FFFF
  );
type
  PWord = ^Word;
begin
  FillChar (Disk, SizeOf (Disk), 0);
  Move (@BootSec^, Disk [0], SizeOf (TSector));
  if Is40 or IsDZ
  then TotalClusters := 800
  else TotalClusters := 1600;
  PWord (@DiskBytes [$13])^ := TotalClusters;
  Move (FAT, Disk [1], SizeOf (FAT));
  Move (FAT, Disk [6], SizeOf (FAT));
end;

function Cluster2FAT (Cluster: Word): Word;
begin
  Cluster2FAT := Cluster*3 div 2;
end;

procedure WriteFAT (Cluster, Next: Word);

  procedure L (Sector: Word);
  var
    Data: ^Word;
  begin
    Data := @DiskBytes [Sector*SizeOf (TSector) + Cluster2FAT (Cluster)];
    if Cluster and 1 = 0
    then Data^ := Data^ and not $0FFF or Next
    else Data^ := Data^ and not $FFF0 or (Next shl 4);
  end;

begin
  L (1);
  L (6);
end;

procedure UpdateFAT;

  function Next (Cluster: Word): Word;
  var
    Data: ^Word;
  begin
    Data := @DiskBytes [1*SizeOf (TSector) + Cluster2FAT (Cluster)];
    if Odd (Cluster)
    then Next := (Data^ shr 4) and $0FFF
    else Next := Data^ and $0FFF;
  end;

const
  BAD = $FF7;
var
  Cluster: Word;
  FreeStart, MaxFreeStart, MaxFreeLength: Word;
  Temp: Word;
begin
  MaxFreeLength := 0;
  FreeStart := 0;
  for Cluster := 2 to TotalClusters+1 do
  begin
    if Cluster < TotalClusters
    then Temp := Next (Cluster)
    else Temp := 1;
    if (Temp = 0) or (Temp = BAD)
    then
    begin
      WriteFAT (Cluster, BAD);
      if FreeStart = 0 then FreeStart := Cluster;
    end
    else
      if FreeStart <> 0 then
      begin
        if Cluster-FreeStart > MaxFreeLength then
        begin
          MaxFreeStart := FreeStart;
          MaxFreeLength := Cluster-FreeStart;
        end;
        FreeStart := 0;
      end;
  end;
  if MaxFreeStart <> 0 then
    for Cluster := MaxFreeStart to MaxFreeStart+MaxFreeLength-1 do
      WriteFAT (Cluster, 0);
end;

var
  DirCount: Integer;

procedure WriteDirEntry (const DirEntry: TDirEntry);
begin
  Move (DirEntry,
    DiskBytes [11*SizeOf (TSector)+DirCount*SizeOf (TDirEntry)],
    SizeOf (TDirEntry));
  Inc (DirCount);
end;

type
  TDiskLabel = array [1..12] of Char;
var
  DiskLabel: TDiskLabel;

procedure ReadLabel;
var
  Sector: TSector;
begin
  if BIOSReadWrite (2, 1, 1, Sector) = 0
  then
    Move (Sector [$200-4-SizeOf (DiskLabel)*3],
      DiskLabel, SizeOf (DiskLabel))
  else
    DiskLabel := 'READ ERROR  ';
end;

var
  Segment: TSegment;
  DirEntry: TDirEntry;

const
  SegmentEnd = $800;
  Permanent = $400;
  Free = $200;
  Protected = $8000;
  RAD50TAB: array [0..40-1] of Char =
    ' ABCDEFGHIJKLMNOPQRSTUVWXYZ$./0123456789';

procedure ReadDir;

  function RAD50 (W: Word): string;
  var
    Result: string [3];
  begin
    Result [3] := RAD50TAB [W mod 40];  W := W div 40;
    Result [2] := RAD50TAB [W mod 40];  W := W div 40;
    Result [1] := RAD50TAB [W];
    Result [0] := #3;
    RAD50 := Result;
  end;

var
  I, J: Integer;
  S: string;
  BlockNo: Word;
  Cluster: Word;
const
  FATEOF = $0FFF;
begin
  DirCount := 0;
  Cluster := 2;
  FillChar (DirEntry, SizeOf (DirEntry), 0);
  Move (DiskLabel, DirEntry, SizeOf (DiskLabel));
  DirEntry.Attr := VolumeID;
  WriteDirEntry (DirEntry);
  BlockNo := 6;
  repeat
    if BIOSReadWrite (2, 2, BlockNo, Segment) <> 0 then
    begin
      {Write (#7);}
      Exit;
    end;
    with Segment do
    begin
      if BlockNo = 6 then
      begin
        RTStartData := StartData;
        RTSegments := Segments;
      end;
      for I := Low (Dir) to High (Dir) do
        with Dir [I] do
          if Attr and SegmentEnd <> 0
          then
            Break
          else
          begin
            if Attr and Permanent <> 0
            then
            begin
              FillChar (DirEntry, SizeOf (DirEntry), 0);
              FillChar (DirEntry, SizeOf (DirEntry.Name)+SizeOf (DirEntry.Ext), ' ');
              S := RAD50 (NA) + RAD50 (ME);
              Move (S [1], DirEntry.Name, Length (S));
              S := RAD50 (Ext);
              Move  (S [1], DirEntry.Ext, Length (S));
              DirEntry.Attr := Archive;
              if Attr and Protected <> 0 then
                DirEntry.Attr := DirEntry.Attr or ReadOnly;
              DirEntry.Size := LongMul (Size, $200);
              DirEntry.Cluster := Cluster;

{
              15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
              || || mm mm mm mm dd dd dd dd dd yy yy yy yy yy
              yy yy yy yy yy yy yy mm mm mm mm dd dd dd dd dd
}
              DirEntry.Date :=
                (Date        and $001F +72 -80) shl 9 +
                (Date shr 10 and $0F          ) shl 5 +
                (Date shr  5 and $1F);

              if Cluster-2+RTStartData+Size > TotalClusters then Exit;
              WriteDirEntry (DirEntry);
              for J := 0 to Integer (Size)-1 -1 do
                WriteFAT (Cluster+J, Cluster+J+1);
              if Size <> 0 then
                WriteFAT (Cluster+Size-1, FATEOF);
            end;
            Inc (Cluster, Size);
          end;
      if Next = 0 then Break;
      BlockNo := 4 + Next*2;
    end;
  until False;
end;

const
  FirstTime: Boolean = True;
  TrapAccess: Boolean = False;
  ForcedChange: Boolean = False;

procedure DiskChanged (Flags, CS, IP, AX, BX, CX, DX, SI, DI, DS, ES, BP: Word); interrupt;

  function ThisIsRTDisk: Boolean;
  const
    NOP = (2*8+4)*8+0;  { 240 }
  var
    PNOP: ^Word;
    Sector: TSector;
  begin
    ThisIsRTDisk := False;
    if BiosReadWrite (2, 1, 0, Sector) = 0 then
    begin
      PNOP := @Sector [0];
      ThisIsRTDisk := PNOP^ = NOP;
    end;
  end;

begin
  TrapAccess := False;
  if ThisIsRTDisk then
  begin
    InitDisk;
    ReadLabel;
    ReadDir;
    UpdateFAT;
    TrapAccess := True;
  end;
end;

var
  O13: procedure;
  _CH, _DH, _CL, _AL: Byte;
  _BX: Word;

procedure WriteDirectoryPostProcess;

  procedure UpdateDir;
  var
    List: TRTCollection;

    function RAD50 (S: PChar): Word;

      function N (C: Char): Byte;
      var
        I: Integer;
      begin
        for I := 0 to High (RAD50TAB) do
          if C = RAD50TAB [I] then
          begin
            N := I;
            Exit;
          end;
        N := 24{X};
      end;

    begin
      RAD50 := (N (S [0])*40+N (S [1]))*40+N (S [2]);
    end;

  var
    SegNo: Integer;
    FileCount: Integer;
    CurBlock: Word;

    procedure IfHole (Block, Size: Word);
    var
      RTDirEntry: PRTDirEntry2;
    begin
      if CurBlock <> Block then
      begin
        GetMem (RTDirEntry, SizeOf (TRTDirEntry2));
        FillChar (RTDirEntry^, SizeOf (RTDirEntry^), 0);
        RTDirEntry^.Attr := Free;
        RTDirEntry^.Block := CurBlock;
        RTDirEntry^.Size := Block-CurBlock;
        List.Insert (RTDirEntry);
      end;
      CurBlock := Block+Size;
    end;

    procedure DoWrite (var RTDirEntry: TRTDirEntry2);
    begin
      if FileCount = 0 then
      begin
        Segment.Segments := RTSegments;
        Segment.StartData := RTDirEntry.Block;
      end;
      Move (RTDirEntry, Segment.Dir [FileCount], SizeOf (TRTDirEntry));
      Inc (FileCount);
      if FileCount = 72-1 then
      begin
        Segment.Dir [FileCount].Attr := SegmentEnd;
        if List.Count > (1+SegNo)*(72-1) then Segment.Next := 1+SegNo+1;
        BIOSReadWrite (3, 2, 6+SegNo*2, Segment);  Inc (SegNo);
        FillChar (Segment, SizeOf (Segment), 0);   FileCount := 0;
      end;
    end;

  var
    W: Word;
    RTDirEntry: PRTDirEntry2;
    DirEntry: PDirEntry;
    DirEntries: Integer;
  begin
    List.Init (0, 1);
    List.Duplicates := True;
    CurBlock := RTStartData;
    DirEntry := @DiskBytes [11*SizeOf (TSector)];
    DirEntries := 0;
    while (DirEntries < 224) and (DirEntry^.Name [1] <> #$00) do
    begin
      if
        (DirEntry^.Attr and VolumeID = 0) and
        (DirEntry^.Name [1] <> #$E5) then
      begin
        GetMem (RTDirEntry, SizeOf (TRTDirEntry2));
        FillChar (RTDirEntry^, SizeOf (RTDirEntry^), 0);
        with RTDirEntry^ do
        begin
          if DirEntry^.Attr and ReadOnly <> 0 then Attr := Protected;
          Attr := Attr or Permanent;
          NA := RAD50 (@DirEntry^.Name [1]);
          ME := RAD50 (@DirEntry^.Name [4]);
          Ext := RAD50 (@DirEntry^.Ext);
          Size := (DirEntry^.Size+SizeOf (TSector)-1) div SizeOf (TSector);
          W := DirEntry^.Date;
          Date :=
            (W shr 9 and $1F            +80 -72) or
            (W       and $1F shl  5) or
            (W shr 5 and $0F shl 10);
{
              15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
              || || mm mm mm mm dd dd dd dd dd yy yy yy yy yy
              yy yy yy yy yy yy yy mm mm mm mm dd dd dd dd dd
}
          if DirEntry^.Cluster <> 0
          then
          begin
            Block := DirEntry^.Cluster - 2 + RTStartData;
            IfHole (Block, Size);
          end
          else
            Block := CurBlock;
        end;
        List.Insert (RTDirEntry);
      end;
      Inc (PtrRec (DirEntry).Ofs, SizeOf (TDirEntry));
      Inc (DirEntries);
    end;
    IfHole (TotalClusters, 0);
    SegNo := 0;
    FileCount := 0;
    FillChar (Segment, SizeOf (Segment), 0);
    TrapAccess := False;
    List.ForEach (@DoWrite);
    if (FileCount <> 0) or  (FileCount = 0) and (SegNo = 0) then
    begin
      Segment.Dir [FileCount].Attr := SegmentEnd;
      BIOSReadWrite (3, 2, 6+SegNo*2, Segment);
    end;
    TrapAccess := True;
    List.Done;
  end;

begin
  UpdateFAT;
  UpdateDir;
  ForcedChange := True;
end;

procedure NonStandard13 (Flags, CS, IP, AX, BX, CX, DX, SI, DI, DS, ES, BP: Word); interrupt;

  function Valid (ByteArray: PByteArray; AL: Byte): Boolean;
  var
    DirEntry: PDirEntry;
    I: Integer;
  begin
    for I := 0 to $10*AL-1 do
    begin
      DirEntry := @ByteArray^ [I*SizeOf (TDirEntry)];
      if DirEntry^.Attr and (Directory or Hidden or SysFile) <> 0 then
      begin
        Valid := False;
        Exit;
      end;
    end;
    Valid := True;
  end;

type
  LH = record L, H: Byte; end;
var
  C: LH absolute CX;
  D: LH absolute DX;
  A: LH absolute AX;
  Cluster: Word;
begin
  Flags := Flags and not 1;
  Cluster := C.H*20 + D.H*10 + C.L - 1;
  if Cluster < StartData
  then
    if Cluster + A.L > StartData
    then
    begin
      A.H := 3;
      Flags := Flags or 1;
    end
    else
    begin
      if A.H = 2
      then
        Move (Disk [Cluster], Ptr (ES, BX)^, SizeOf (TSector)*A.L)
      else
      begin
        if Cluster in [11..24]
        then
          if Valid (Ptr (ES, BX), A.L)
          then
          begin
            Move (Ptr (ES, BX)^, Disk [Cluster], SizeOf (TSector)*A.L);
            WriteDirectoryPostProcess;
          end
          else
          begin
            A.H := 3;
            Flags := Flags or 1;
          end
        else
          Move (Ptr (ES, BX)^, Disk [Cluster], SizeOf (TSector)*A.L);
      end;
    end
  else
  begin
    Inc (Cluster, -StartData+RTStartData);
    TrapAccess := False;
    A.H := Byte (BIOSReadWrite (A.H, A.L, Cluster, Ptr (ES, BX)^));
    TrapAccess := True;
    if A.H <> 0 then Flags := Flags or 1;
  end;
end;

const
  TempStackSize = $200;
var
  _OldSS, _OldSP,
  _SS: Word;
  TempStack: array [1..TempStackSize] of Word;

procedure N13; assembler;
asm
  push ds

  push SEG @Data
  pop  ds

  cmp  ax, 02FFh
  jne  @@nIs

  mov  Is40, cl
  mov  IsDZ, ch
  mov  FirstTime, True
  jmp  @@ret

@@nIs:
  cmp  ah, 16h
  jne  @@ntc
  or   dl, dl
  jnz  @@std

  cmp  ForcedChange, False
  jz   @@nfc

  mov  ForcedChange, False

  mov  ah, 06
  stc

  jmp  @@ret

@@nfc:
  mov  _OldSS, ss
  mov  _OldSP, sp
  mov  _SS, ds
  cli
  mov  ss, _SS
  mov  sp, OFFSET TempStack + TempStackSize*2
  sti

  pushf
  call O13

  pushf
  cmp  FirstTime, False
  jne  @@FirstTime

  or   ah, ah
  je   @@nc

@@FirstTime:
  mov  FirstTime, False

  pushf
  call FAR PTR DiskChanged

@@nc:
  popf

  cli
  mov  ss, _OldSS
  mov  sp, _OldSP
  sti

  jmp  @@ret

@@ntc:
  cmp  ah, 02
  je   @@nstd
  cmp  ah, 03
  jne  @@std
@@nstd:
  or   dl, dl
  jnz  @@std
  cmp  TrapAccess, 0
  je   @@std

  push bx
  push cx
  push dx

  pushf
  call FAR PTR NonStandard13

@@err:
@@nsok:
  pop  dx
  pop  cx
  pop  bx

  jmp  @@ret

@@std:

  pushf
  call O13

@@ret:
  pop  ds
  retf 2
end;

var
  CFG: string;

procedure ForHelp;
begin
  WriteLn ('For help, type PAFRT ?');
end;

procedure Option;
var
  I: Integer;
begin
  if ParamCount = 1
  then
  begin
    CFG := ParamStr (1);
    for I := 1 to Length (CFG) do
      CFG [I] := UpCase (CFG [I]);
    if (CFG = 'MY') or (CFG = 'MY80')
    then
    begin
      Is40 := False;
      IsDZ := False;
    end
    else
      if CFG = 'MY40'
      then
      begin
        Is40 := True;
        IsDZ := False;
      end
      else
        if CFG = 'DZ'
        then
        begin
          Is40 := False;
          IsDZ := True;
        end
        else
          if CFG = '?'
          then
          begin
            WriteLn;
            WriteLn ('Usage: PAFRT [parameter]');
            WriteLn;
            WriteLn ('Parameters:');
            WriteLn ('   MY80 = MY');
            WriteLn ('   MY40');
            WriteLn ('   DZ');
            Halt (0);
          end
          else
          begin
            WriteLn ('Illegal parameter: "', CFG, '"');
            ForHelp; Halt (1);
          end;
  end
  else
    if ParamCount = 0
    then
    begin
      WriteLn ('Parameter missing');
      ForHelp; Halt (1);
    end
    else
    begin
      WriteLn ('Too many parameters');
      ForHelp; Halt (1);
    end;
end;

{$IFDEF Debug}
var
  OldExitProc: Pointer;
  Command: string;

function Parameters: string;
var
  I: Integer;
  Result: string;
begin
  Result := '';
  for I := 1 to ParamCount do
    Result := Result + ParamStr (I) + ' ';
  Parameters := Result;
end;

procedure PAFExitProc;
begin
  ExitProc := OldExitProc;
  SetIntVec ($13, @O13);
end;
{$ENDIF}

begin
  WriteLn ('The RT-IBM disk emulator  Version 4.1  Copyright 1994 (c) Alexander Petrosyan');
{$IFDEF Debug}
  WriteLn ('DEBUG Version.  a lot of BIG BUGs are hiding here  do not copy!');
  if ParamCount = 0 then
  begin
    WriteLn (#7'Not installed');
    WriteLn;
    WriteLn ('PAFRT <command to run>');
    Halt (1);
  end;
{$ELSE}
  Option;
{$ENDIF}

  InitDisk;
  GetIntVec ($13, @O13);
  if Ofs (O13) = Ofs (N13)
  then WriteLn ('Already installed')
  else WriteLn ('Installed');
  WriteLn;
  WriteLn ('Now you can work with ', CFG, ' disks as if they are IBM disks');
  if Ofs (O13) = Ofs (N13) then
  begin
    asm
      mov  ax, 02FFh
      mov  cl, Is40
      mov  ch, IsDZ
      int  13h
    end;
    Halt (1);
  end;
  SetIntVec ($13, @N13);

{$IFDEF Debug}
  OldExitProc := ExitProc;  ExitProc := @PAFExitProc;
  Command := Parameters;
  WriteLn;
  WriteLn (Command);
  SwapVectors;
  Exec (GetEnv ('comspec'), '/c '+Command);
  SwapVectors;
  SetIntVec ($13, @O13);
{$ELSE}
  Keep (0);
{$ENDIF}
end.
